#include<errno.h>
#include<string.h>
#include<sys/stat.h>
#include<unistd.h>
#include<stdio.h>
#include<sys/types.h>
#include<fcntl.h>
#include<stdlib.h>
#include<assert.h>

#include"stree.h"
#include"file_hdl.h"

/*
  Tiedostossa oleva transition-tietue.

  offset len
    0     1   byte transId
    1     4   txt_p begin (byte4)
    5     4   txt_p end (byte4)
    9     1   byte n_trans
    10    4   idx_p childs  (byte4)
    14

  Kun n_trans==0 on childs osoittimessa ptteen alkukohta.
*/

#define TID_LEN 1
#define BEGIN_LEN 4
#define END_LEN 4
#define NTRANS_LEN 1
#define CHILDP_LEN 4

#define TID_OFF 0
#define BEGIN_OFF 1
#define END_OFF 5
#define NTRANS_OFF 9
#define CHILDP_OFF 10



int load_idx_header(st_meta *data, struct stat *txtstat) 
{

  long tmp4;

  if(read(data->idx_filedes,&tmp4,4)<0) {
    return -1;
  }
  if(tmp4!=IDX_MAGIC) {
    fprintf(stderr,"WRONG MAGIC NUMBER!!\n");
    close(data->idx_filedes);
    return -1;
  }

  if(read(data->idx_filedes,&tmp4,4)<0) {
    return -1;
  }
  data->fsize=tmp4;

  if(read(data->idx_filedes,&tmp4,4)<0) {
    return -1;
  }
  data->fmodified=tmp4;

  if(read(data->idx_filedes,data->fname,256)<0) {
    return -1;
  }

  if(read(data->idx_filedes,&data->text_p_size,1)<0) {
    return -1;
  }
  if(read(data->idx_filedes,&data->with_sfx_links,1)<0) {
    return -1;
  }


  data->root.trans=NULL;
  data->root.sfx_link=NULL;

  if(read(data->idx_filedes,&data->root.disk_trans,4)<0) {
    return -1;
  }
  if(read(data->idx_filedes,&data->root.n_tran,1)<0) {
    return -1;
  }

  return 0;
}

/* Checks that transitions of node here is in RAM. */
int valid_transitions(st_meta *data,st_node *here)
{
  int pos,i;
  st_trans *ret_buf;
  char *dbuf,nchildren;

  if(here->trans!=NULL)
    return 1;
  if(data->use_mode=='c') { /* No updates */
    return 1;
  }

  pos=lseek(data->idx_filedes,here->disk_trans,SEEK_SET);

  if(pos!=here->disk_trans) {
    return -1;
  }


  ret_buf=malloc(sizeof(st_trans)*here->n_tran);
  if(ret_buf==NULL)
    mem_error();

  memset(ret_buf,0,sizeof(st_trans)*here->n_tran);

  dbuf=malloc(TRANSITION_SIZE*here->n_tran);
  if(dbuf==NULL)
    mem_error();


  if(read(data->idx_filedes,dbuf,TRANSITION_SIZE*here->n_tran)<0)
    return -1;
  

  for(i=0;i<here->n_tran;i++) {
    memcpy(&ret_buf[i].transId,dbuf+i*TRANSITION_SIZE+TID_OFF,TID_LEN);
    memcpy(&ret_buf[i].begin,dbuf+i*TRANSITION_SIZE+BEGIN_OFF,BEGIN_LEN);
    memcpy(&ret_buf[i].end,dbuf+i*TRANSITION_SIZE+END_OFF,END_LEN);

    memcpy(&nchildren,dbuf+i*TRANSITION_SIZE+NTRANS_OFF,NTRANS_LEN);

    if(nchildren>0) {
      ret_buf[i].childs=malloc(sizeof(st_node));
      if(ret_buf[i].childs==NULL)
	mem_error();
      memset(ret_buf[i].childs,0,sizeof(st_node));
      ret_buf[i].childs->n_tran=nchildren;
      memcpy(&ret_buf[i].childs->disk_trans,dbuf+i*TRANSITION_SIZE+
	     CHILDP_OFF,CHILDP_LEN);
      ret_buf[i].childs->trans=NULL;
    } else {
      long sfxbgin;

      memcpy(&sfxbgin,dbuf+i*TRANSITION_SIZE+CHILDP_OFF,CHILDP_LEN);
      assert(ret_buf[i].end==INFTY_P);
      ret_buf[i].end=-sfxbgin-1;
    } 
  }


  free(dbuf);

  here->trans=ret_buf;


  return 1;
}

int init_files(st_meta *data)
{
  struct stat txt_f;
  int tmp;
  char idx_fname[256];

  if(stat(data->fname,&txt_f)==-1) {
    fprintf(stderr,"Textfile: %s\n",strerror(errno));
    return -1;
  }
  data->txt_filedes=open(data->fname,O_RDONLY);
  if(data->txt_filedes==-1) {
    fprintf(stderr,"Textfile open: %s\n",strerror(errno));
    return -1;
  }
  
  if(data->use_mode=='f') {
    strcpy(idx_fname,data->fname);
    strncat(idx_fname,".sxt",255-strlen(data->fname));

    data->idx_filedes=open(idx_fname,O_RDONLY);
    if(data->idx_filedes==-1) {
      fprintf(stderr,"No index created for %s\n",data->fname);
      /*fprintf(stderr,"Index open: %s\n",strerror(errno));*/
      return -1;
    }
    tmp=load_idx_header(data, &txt_f);
    
    if(tmp!=-1 && 
       (data->fsize!=txt_f.st_size || data->fmodified!=txt_f.st_mtime)) {
      fprintf(stderr,"WARNING: Indexed file modified after index creation!\n");
    }
    return tmp;

  } else if(data->use_mode=='c') {
    strcpy(idx_fname,data->fname);
    strncat(idx_fname,".sxt",255-strlen(data->fname));

    data->idx_filedes=open(idx_fname,O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP);
    if(data->idx_filedes==-1) {
      fprintf(stderr,"Index open: %s\n",strerror(errno));
      return -1;
    }
    data->fsize=txt_f.st_size;
    data->fmodified=txt_f.st_mtime;
  } else {
    return -1;
  }
  return 0;
}

/* Header:

   byte4 magic =140478
   byte4 fsize
   byte4 fmodified
   byte256 fname
   byte text_p_size
   byte with_sfx_links
   byte4 root_p
   byte rtrans_count;
*/
#define HEADER_SIZE 275

/* Save the address and size of the root node to file header. */
int save_root(st_meta *data,long root_p,unsigned char count)
{
  if(lseek(data->idx_filedes,HEADER_SIZE-5,SEEK_SET)!=(HEADER_SIZE-5)) {
    fprintf(stderr,"Index Seeking failed! %s \n",strerror(errno));
  }
  if(write(data->idx_filedes,&root_p,4)<0)
    return -1;
  if(write(data->idx_filedes,&count,1)<0)
    return -1;

  return 5;
}


int save_header(st_meta *data)
{
  long tmp4=IDX_MAGIC;
  char tmp1;

  if(write(data->idx_filedes,&tmp4,4)<0)
    return -1;

  tmp4=data->fsize;
  if(write(data->idx_filedes,&tmp4,4)<0)
    return -1;

  tmp4=data->fmodified;
  if(write(data->idx_filedes,&tmp4,4)<0)
    return -1;

  if(write(data->idx_filedes,data->fname,256)<0)
    return -1;

  tmp1=data->text_p_size;
  if(write(data->idx_filedes,&tmp1,1)<0)
    return -1;
  
  tmp1=data->with_sfx_links;
  if(write(data->idx_filedes,&tmp1,1)<0)
    return -1;
  
  tmp4=0;
  if(write(data->idx_filedes,&tmp4,4)<0)
    return -1;
  tmp1=0;
  if(write(data->idx_filedes,&tmp1,1)<0)
    return -1;
  
  return HEADER_SIZE;
}



/* We don't want to stack these pointers. */
static st_meta *meta_for_save;

int save_nodes(st_node *this,long dist)
{
  char *data=malloc(TRANSITION_SIZE*this->n_tran);
  int i;
  long my_pos,tmp;

  if(data==NULL) {
    mem_error();
  }

  memset(data,0,TRANSITION_SIZE*this->n_tran);

  for(i=0;i<this->n_tran;i++) {
    memcpy(data+i*TRANSITION_SIZE+TID_OFF,&this->trans[i].transId,TID_LEN);
    memcpy(data+i*TRANSITION_SIZE+BEGIN_OFF,&this->trans[i].begin,BEGIN_LEN);
    memcpy(data+i*TRANSITION_SIZE+END_OFF,&this->trans[i].end,END_LEN);
    if(this->trans[i].childs!=NULL) {
      memcpy(data+i*TRANSITION_SIZE+NTRANS_OFF,&this->trans[i].childs->n_tran,NTRANS_LEN);
      tmp=save_nodes(this->trans[i].childs,
		     dist+this->trans[i].end-this->trans[i].begin+1);
      memcpy(data+i*TRANSITION_SIZE+CHILDP_OFF,&tmp,CHILDP_LEN);
    } else {
      /* The beginning of the suffix */
      tmp=this->trans[i].begin-dist;
      memcpy(data+i*TRANSITION_SIZE+CHILDP_OFF,&tmp,CHILDP_LEN);
    }
  }

  my_pos=lseek(meta_for_save->idx_filedes,0,SEEK_CUR);

  if(write(meta_for_save->idx_filedes,data,TRANSITION_SIZE*this->n_tran)<0)
    return -1;

  free(data);

  return my_pos;
}


int save_tree(st_meta *data)
{
  long pointer;

  if(save_header(data)<0) {
    fprintf(stderr,"Indexfile header save failed! %s\n",strerror(errno));
  }
  meta_for_save=data;

  pointer=save_nodes(&(data->root),0);

  if(pointer<0) {
    fprintf(stderr,"Indexfile data save failed! %s\n",strerror(errno));
  }

  if(save_root(data,pointer,data->root.n_tran)<0) {
    fprintf(stderr,"Indexfile save failed! %s\n",strerror(errno));
  }

  return 0;

}

/* Read a new block of text to cache. It's a terrible waiste of time
   to ask single character from OS one at a time. 

   lowdata contains text t:  lowstart  <= t < highstart
   highdata contains text t: highstart <= t < highlimit

   cache must be updated if t >= highlimit
*/

#define CACHEBLOCK 512
void read_cache(st_meta *data, long k) 
{
  char *swdata;

  if(data->cache.lowdata==NULL) {
    data->cache.lowdata=malloc(CACHEBLOCK);
    if(data->cache.lowdata==NULL) {
      mem_error();
    }
  }
  swdata=data->cache.lowdata;
  data->cache.lowdata=data->cache.highdata;
  data->cache.lowstart=data->cache.highstart;

  data->cache.highdata=swdata;
  data->cache.highstart=data->cache.highlimit;

  data->cache.highlimit+=CACHEBLOCK;
  if(data->cache.highlimit>=data->fsize) {
    data->cache.highlimit=data->fsize-1;
  }

  read_string(data,data->cache.highdata,data->cache.highstart,
	      data->cache.highlimit-1,CACHEBLOCK-1);


}

char read_tchar(st_meta *data, long k)
{
  int tmp;
  int result;

  data->cache.uses++;
  
  if(k>=data->cache.highlimit) {
    data->cache.highmiss++;
    read_cache(data,k);
  }

  if(k>=data->cache.highstart) {
    return data->cache.highdata[k-data->cache.highstart];
  } else if(k>=data->cache.lowstart) {
    return data->cache.lowdata[k-data->cache.lowstart];
  }
  
  data->cache.lowmiss++;

  if(lseek(data->txt_filedes,k,SEEK_SET)!=k) {
    fprintf(stderr,"Textfile Seeking failed!\n");
  }

  if(k>=data->fsize)
    return 0;

  result=read(data->txt_filedes,(void*)&tmp,1);

  if(result==-1) {
    fprintf(stderr,"Reading textfile: %s\n",strerror(errno));
    abort();
  }

  return (char)tmp;
}

/* Reads a text t where:  bgin<= t <=end  and end-bgin<=limit */
char *read_string(st_meta *data, char *buf, long bgin, long end, long limit)
{
  long length=end-bgin+1,numread;

  end=(length>limit?bgin+limit:end);
  length=end-bgin+1;

  if(bgin!=lseek(data->txt_filedes,bgin,SEEK_SET)) {
    fprintf(stderr,"Textfile Seeking failed!\n");
  }
  

  numread=read(data->txt_filedes,buf,length);
  
  buf[numread]=0;

  return buf;
}
